/*
 * Copyright 2024 Vsevolod Stakhov
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef RSPAMD_RSPAMD_SIMDUTF_H
#define RSPAMD_RSPAMD_SIMDUTF_H

#pragma once
#include "config.h"

#ifdef __cplusplus
extern "C" {
#endif

void rspamd_fast_utf8_library_init(unsigned flags);
const char *rspamd_fast_utf8_library_impl_name(void);
off_t rspamd_fast_utf8_validate(const unsigned char *data, size_t len);
off_t rspamd_fast_utf8_validate_ref(const unsigned char *data, size_t len);

#ifdef __cplusplus
}
#endif

#endif//RSPAMD_RSPAMD_SIMDUTF_H
